@@ -1,14 +1,16 @@
 import copy
+import os
 import warnings
 import pandas as pd
 import numpy as np
 import seaborn as sns
 import matplotlib.pyplot as plt
+from pathlib import Path
 from tqdm import tqdm
 from scipy import stats
 from typing import Tuple, Dict, Union
 from scipy.spatial.distance import cdist
-from sklearn.model_selection import train_test_split
+from sklearn.model_selection import KFold
 from sklearn.tree import DecisionTreeClassifier
 from sklearn.neural_network import MLPClassifier
 from sklearn.decomposition import PCA
@@ -19,6 +21,8 @@ from sklearn.linear_model import Lasso, Ridge, ElasticNet, LogisticRegression
 from dython.nominal import compute_associations, numerical_encoding
 from .viz import *
 from .metrics import *
+from .notebook import visualize_notebook, isnotebook, EvaluationResult
+from .utils import dict_to_df
 
 
 class TableEvaluator:
@@ -28,8 +32,7 @@ class TableEvaluator:
     """
 
     def __init__(self, real: pd.DataFrame, fake: pd.DataFrame, cat_cols=None, unique_thresh=0, metric='pearsonr',
-                 verbose=False, n_samples=None,
-                 name: str = None, seed=1337):
+                 verbose=False, n_samples=None, name: str = None, seed=1337):
         """
         :param real: Real dataset (pd.DataFrame)
         :param fake: Synthetic dataset (pd.DataFrame)
@@ -86,16 +89,18 @@ class TableEvaluator:
         self.fake.loc[:, self.numerical_columns] = self.fake.loc[:, self.numerical_columns].fillna(
             self.fake[self.numerical_columns].mean())
 
-    def plot_mean_std(self):
+    def plot_mean_std(self, fname=None):
         """
         Class wrapper function for plotting the mean and std using `viz.plot_mean_std`.
+        :param fname: If not none, saves the plot with this file name. 
         """
-        plot_mean_std(self.real, self.fake)
+        plot_mean_std(self.real, self.fake, fname=fname)
 
-    def plot_cumsums(self, nr_cols=4):
+    def plot_cumsums(self, nr_cols=4, fname=None):
         """
         Plot the cumulative sums for all columns in the real and fake dataset. Height of each row scales with the length of the labels. Each plot contains the
         values of a real columns and the corresponding fake column.
+        :param fname: If not none, saves the plot with this file name. 
         """
         nr_charts = len(self.real.columns)
         nr_rows = max(1, nr_charts // nr_cols)
@@ -118,12 +123,17 @@ class TableEvaluator:
             f = self.fake.iloc[:, self.real.columns.tolist().index(col)]
             cdf(r, f, col, 'Cumsum', ax=axes[i])
         plt.tight_layout(rect=[0, 0.02, 1, 0.98])
+
+        if fname is not None: 
+            plt.savefig(fname)
+
         plt.show()
 
-    def plot_distributions(self, nr_cols=3):
+    def plot_distributions(self, nr_cols=3, fname=None):
         """
         Plot the distribution plots for all columns in the real and fake dataset. Height of each row of plots scales with the length of the labels. Each plot
         contains the values of a real columns and the corresponding fake column.
+        :param fname: If not none, saves the plot with this file name. 
         """
         nr_charts = len(self.real.columns)
         nr_rows = max(1, nr_charts // nr_cols)
@@ -143,15 +153,9 @@ class TableEvaluator:
         axes = ax.flatten()
         for i, col in enumerate(self.real.columns):
             if col not in self.categorical_columns:
-                try:
-                    sns.distplot(self.real[col], ax=axes[i], label='Real')
-                    sns.distplot(self.fake[col], ax=axes[i], color='darkorange', label='Fake')
-                except RuntimeError:
-                    axes[i].clear()
-                    sns.distplot(self.real[col], ax=axes[i], label='Real', kde=False)
-                    sns.distplot(self.fake[col], ax=axes[i], color='darkorange', label='Fake', kde=False)
+                plot_df = pd.DataFrame({col: self.real[col].append(self.fake[col]), 'kind': ['real'] * self.n_samples + ['fake'] * self.n_samples})
+                fig = sns.histplot(plot_df, x=col, hue='kind', ax=axes[i], stat='probability', legend=True)
                 axes[i].set_autoscaley_on(True)
-                axes[i].legend()
             else:
                 real = self.real.copy()
                 fake = self.fake.copy()
@@ -170,16 +174,21 @@ class TableEvaluator:
                       .pipe((sns.barplot, "data"), x=x, y=y, hue=hue, ax=axes[i], saturation=0.8, palette=palette))
                 ax.set_xticklabels(axes[i].get_xticklabels(), rotation='vertical')
         plt.tight_layout(rect=[0, 0.02, 1, 0.98])
+
+        if fname is not None: 
+            plt.savefig(fname)
+
         plt.show()
 
-    def plot_correlation_difference(self, plot_diff=True, **kwargs):
+    def plot_correlation_difference(self, plot_diff=True, fname=None, **kwargs):
         """
         Plot the association matrices for each table and, if chosen, the difference between them.
 
         :param plot_diff: whether to plot the difference
+        :param fname: If not none, saves the plot with this file name.
         :param kwargs: kwargs for sns.heatmap
         """
-        plot_correlation_difference(self.real, self.fake, cat_cols=self.categorical_columns, plot_diff=plot_diff,
+        plot_correlation_difference(self.real, self.fake, cat_cols=self.categorical_columns, plot_diff=plot_diff, fname=fname,
                                     **kwargs)
 
     def correlation_distance(self, how: str = 'euclidean') -> float:
@@ -212,12 +221,13 @@ class TableEvaluator:
             fake_corr.values
         )
 
-    def plot_pca(self):
+    def plot_pca(self, fname=None):
         """
         Plot the first two components of a PCA of real and fake data.
+        :param fname: If not none, saves the plot with this file name.
         """
-        real = numerical_encoding(self.real, nominal_columns=self.categorical_columns)
-        fake = numerical_encoding(self.fake, nominal_columns=self.categorical_columns)
+        real, fake = self.convert_numerical()
+
         pca_r = PCA(n_components=2)
         pca_f = PCA(n_components=2)
 
@@ -230,6 +240,10 @@ class TableEvaluator:
         sns.scatterplot(ax=ax[1], x=fake_t[:, 0], y=fake_t[:, 1])
         ax[0].set_title('Real data')
         ax[1].set_title('Fake data')
+
+        if fname is not None: 
+            plt.savefig(fname)
+
         plt.show()
 
     def get_copies(self, return_len: bool = False) -> Union[pd.DataFrame, int]:
@@ -279,11 +293,7 @@ class TableEvaluator:
         self.pca_r = PCA(n_components=5)
         self.pca_f = PCA(n_components=5)
 
-        real = self.real
-        fake = self.fake
-
-        real = numerical_encoding(real, nominal_columns=self.categorical_columns)
-        fake = numerical_encoding(fake, nominal_columns=self.categorical_columns)
+        real, fake = self.convert_numerical()
 
         self.pca_r.fit(real)
         self.pca_f.fit(fake)
@@ -332,10 +342,10 @@ class TableEvaluator:
                                                          [self.real_y_test, self.fake_y_test], ['real', 'fake']):
                     predictions_classifier_real = r_classifier.predict(dataset)
                     predictions_classifier_fake = f_classifier.predict(dataset)
-                    f1_r = f1_score(target, predictions_classifier_real, average="micro")
-                    f1_f = f1_score(target, predictions_classifier_fake, average="micro")
+                    f1_r = f1_score(target, predictions_classifier_real, average='micro')
+                    f1_f = f1_score(target, predictions_classifier_fake, average='micro')
                     jac_sim = jaccard_score(predictions_classifier_real, predictions_classifier_fake, average='micro')
-                    row = {'index': f'{estimator_name}_{dataset_name}_testset', 'f1_real': f1_r, 'f1_fake': f1_f,
+                    row = {'index': f'{estimator_name}_{dataset_name}', 'f1_real': f1_r, 'f1_fake': f1_f,
                            'jaccard_similarity': jac_sim}
                     rows.append(row)
             results = pd.DataFrame(rows).set_index('index')
@@ -354,19 +364,30 @@ class TableEvaluator:
             raise Exception(f'self.target_type should be either \'class\' or \'regr\', but is {self.target_type}.')
         return results
 
-    def visual_evaluation(self, **kwargs):
+    def visual_evaluation(self, save_dir=None, **kwargs):
         """
         Plot all visual evaluation metrics. Includes plotting the mean and standard deviation, cumulative sums, correlation differences and the PCA transform.
-
+        :save_dir: directory path to save images 
         :param kwargs: any kwargs for matplotlib.
         """
-        self.plot_mean_std()
-        self.plot_cumsums()
-        self.plot_distributions()
-        self.plot_correlation_difference(**kwargs)
-        self.plot_pca()
+        if save_dir is None: 
+            self.plot_mean_std()
+            self.plot_cumsums()
+            self.plot_distributions()
+            self.plot_correlation_difference(**kwargs)
+            self.plot_pca()    
+        else: 
+            save_dir = Path(save_dir)
+            save_dir.mkdir(parents=True, exist_ok=True)
 
-    def statistical_evaluation(self) -> float:
+            self.plot_mean_std(fname=save_dir/'mean_std.png')
+            self.plot_cumsums(fname=save_dir/'cumsums.png')
+            self.plot_distributions(fname=save_dir/'distributions.png')
+            self.plot_correlation_difference(fname=save_dir/'correlation_difference.png', **kwargs)
+            self.plot_pca(fname=save_dir/'pca.png') 
+        
+
+    def basic_statistical_evaluation(self) -> float:
         """
         Calculate the correlation coefficient between the basic properties of self.real and self.fake using Spearman's Rho. Spearman's is used because these
         values can differ a lot in magnitude, and Spearman's is more resilient to outliers.
@@ -424,10 +445,25 @@ class TableEvaluator:
         Special function to convert dataset to a numerical representations while making sure they have identical columns. This is sometimes a problem with
         categorical columns with many values or very unbalanced values
 
+        :return: Real and fake dataframe factorized using the pandas function
+        """
+        real = self.real
+        fake = self.fake
+        for c in self.categorical_columns:
+            if real[c].dtype == 'object':
+                real[c] = pd.factorize(real[c], sort=True)[0]
+                fake[c] = pd.factorize(fake[c], sort=True)[0]
+
+        return real, fake
+
+    def convert_numerical_one_hot(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
+        """
+        Special function to convert dataset to a numerical representations while making sure they have identical columns. This is sometimes a problem with
+        categorical columns with many values or very unbalanced values
+
         :return: Real and fake dataframe with categorical columns one-hot encoded and binary columns factorized.
         """
         real = numerical_encoding(self.real, nominal_columns=self.categorical_columns)
-
         columns = sorted(real.columns.tolist())
         real = real[columns]
         fake = numerical_encoding(self.fake, nominal_columns=self.categorical_columns)
@@ -435,9 +471,10 @@ class TableEvaluator:
             if col not in fake.columns.tolist():
                 fake[col] = 0
         fake = fake[columns]
+
         return real, fake
 
-    def estimator_evaluation(self, target_col: str, target_type: str = 'class') -> float:
+    def estimator_evaluation(self, target_col: str, target_type: str = 'class', kfold: bool = False) -> float:
         """
         Method to do full estimator evaluation, including training. And estimator is either a regressor or a classifier, depending on the task. Two sets are
         created of each of the estimators `S_r` and `S_f`, for the real and fake data respectively. `S_f` is trained on ``self.real`` and `S_r` on
@@ -446,43 +483,26 @@ class TableEvaluator:
 
         :param target_col: which column should be considered the target both both the regression and classification task.
         :param target_type: what kind of task this is. Can be either ``class`` or ``regr``.
+        :param kfold: if set to True, will perform 5-fold CV, otherwise will just train on 80% and test on 20% of the data once.
         :return: Correlation value or 1 - MAPE
         """
         self.target_col = target_col
         self.target_type = target_type
 
         # Convert both datasets to numerical representations and split x and  y
-        real_x = numerical_encoding(self.real.drop([target_col], axis=1), nominal_columns=self.categorical_columns)
+        real, fake = self.convert_numerical()
 
-        columns = sorted(real_x.columns.tolist())
-        real_x = real_x[columns]
-        fake_x = numerical_encoding(self.fake.drop([target_col], axis=1), nominal_columns=self.categorical_columns)
-        for col in columns:
-            if col not in fake_x.columns.tolist():
-                fake_x[col] = 0
-        fake_x = fake_x[columns]
+        real_x = real.drop([target_col], axis=1)
+        fake_x = fake.drop([target_col], axis=1)
 
         assert real_x.columns.tolist() == fake_x.columns.tolist(), f'real and fake columns are different: \n{real_x.columns}\n{fake_x.columns}'
 
-        if self.target_type == 'class':
-            # Encode real and fake target the same
-            real_y, uniques = pd.factorize(self.real[target_col])
-            mapping = {key: value for value, key in enumerate(uniques)}
-            fake_y = [mapping.get(key) for key in self.fake[target_col].tolist()]
-        elif self.target_type == 'regr':
-            real_y = self.real[target_col]
-            fake_y = self.fake[target_col]
-        else:
-            raise Exception(f'Target Type must be regr or class')
+        real_y = real[target_col]
+        fake_y = fake[target_col]
 
         # For reproducibilty:
         np.random.seed(self.random_seed)
 
-        self.real_x_train, self.real_x_test, self.real_y_train, self.real_y_test = train_test_split(real_x, real_y,
-                                                                                                    test_size=0.2)
-        self.fake_x_train, self.fake_x_test, self.fake_y_train, self.fake_y_test = train_test_split(fake_x, fake_y,
-                                                                                                    test_size=0.2)
-
         if target_type == 'regr':
             self.estimators = [
                 RandomForestRegressor(n_estimators=20, max_depth=5, random_state=42),
@@ -500,19 +520,40 @@ class TableEvaluator:
         else:
             raise ValueError(f'target_type must be \'regr\' or \'class\'')
 
-        self.r_estimators = copy.deepcopy(self.estimators)
-        self.f_estimators = copy.deepcopy(self.estimators)
         self.estimator_names = [type(clf).__name__ for clf in self.estimators]
 
-        for estimator in self.estimators:
-            assert hasattr(estimator, 'fit')
-            assert hasattr(estimator, 'score')
+        # K Fold
+        kf = KFold(n_splits=5)
+        res = []
+        for train_index, test_index in kf.split(real_y):
+            self.real_x_train = real_x.iloc[train_index]
+            self.real_x_test = real_x.iloc[test_index]
+            self.real_y_train = real_y.iloc[train_index]
+            self.real_y_test = real_y.iloc[test_index]
+            self.fake_x_train = fake_x.iloc[train_index]
+            self.fake_x_test = fake_x.iloc[test_index]
+            self.fake_y_train = fake_y.iloc[train_index]
+            self.fake_y_test = fake_y.iloc[test_index]
 
-        self.fit_estimators()
-        self.estimators_scores = self.score_estimators()
-        print('\nClassifier F1-scores and their Jaccard similarities:') if self.target_type == 'class' \
-            else print('\nRegressor MSE-scores and their Jaccard similarities:')
-        print(self.estimators_scores.to_string())
+            self.r_estimators = copy.deepcopy(self.estimators)
+            self.f_estimators = copy.deepcopy(self.estimators)
+
+            for estimator in self.estimators:
+                assert hasattr(estimator, 'fit')
+                assert hasattr(estimator, 'score')
+
+            self.fit_estimators()
+            res.append(self.score_estimators())
+
+            # Break the loop if we don't want the kfold
+            if not kfold:
+                break
+
+        self.estimators_scores = pd.concat(res).groupby(level=0).mean()
+        if self.verbose:
+            print('\nClassifier F1-scores and their Jaccard similarities:') if self.target_type == 'class' \
+                else print('\nRegressor MSE-scores and their Jaccard similarities:')
+            print(self.estimators_scores.to_string())
 
         if self.target_type == 'regr':
             corr, p = self.comparison_metric(self.estimators_scores['real'], self.estimators_scores['fake'])
@@ -520,8 +561,10 @@ class TableEvaluator:
         elif self.target_type == 'class':
             mean = mean_absolute_percentage_error(self.estimators_scores['f1_real'], self.estimators_scores['f1_fake'])
             return 1 - mean
+        else:
+            raise ValueError('`self.target_type` should be `regr` or `class`.')
 
-    def row_distance(self, n_samples: int = None) -> Tuple[float, float]:
+    def row_distance(self, n_samples: int = None) -> Tuple[np.number, np.number]:
         """
         Calculate mean and standard deviation distances between `self.fake` and `self.real`.
 
@@ -530,8 +573,8 @@ class TableEvaluator:
         """
         if n_samples is None:
             n_samples = len(self.real)
-        real = numerical_encoding(self.real, nominal_columns=self.categorical_columns)
-        fake = numerical_encoding(self.fake, nominal_columns=self.categorical_columns)
+
+        real, fake = self.convert_numerical_one_hot()
 
         columns = sorted(real.columns.tolist())
         real = real[columns]
@@ -559,10 +602,13 @@ class TableEvaluator:
 
         :return: Column correlations between ``self.real`` and ``self.fake``.
         """
-        return column_correlations(self.real, self.fake, self.categorical_columns)
 
-    def evaluate(self, target_col: str, target_type: str = 'class', metric: str = None, verbose=None,
-                 n_samples_distance: int = 20000) -> Dict:
+        real, fake = self.convert_numerical()
+
+        return column_correlations(real, fake, self.categorical_columns)
+
+    def evaluate(self, target_col: str, target_type: str = 'class', metric: str = None, verbose: bool = None,
+                 n_samples_distance: int = 20000, kfold: bool = False, notebook: bool = False, return_outputs: bool = False) -> Dict:
         """
         Determine correlation between attributes from the real and fake dataset using a given metric.
         All metrics from scipy.stats are available.
@@ -572,47 +618,110 @@ class TableEvaluator:
         :param metric: overwrites self.metric. Scoring metric for the attributes.
             By default Pearson's r is used. Alternatives include Spearman rho (scipy.stats.spearmanr) or Kendall Tau (scipy.stats.kendalltau).
         :param n_samples_distance: The number of samples to take for the row distance. See documentation of ``tableEvaluator.row_distance`` for details.
+        :param kfold: Use a 5-fold CV for the ML estimators if set to True. Train/Test on 80%/20% of the data if set to False.
+        :param notebook: Better visualization of the results in a python notebook
         :param verbose: whether to print verbose logging.
+        :param return_outputs: Will omit printing and instead return a dictionairy with all results. 
         """
-        if verbose is not None:
-            self.verbose = verbose
-        if metric is not None:
-            self.comparison_metric = metric
+        self.verbose = verbose if verbose is not None else self.verbose
+        self.comparison_metric = metric if metric is not None else self.comparison_metric
 
         warnings.filterwarnings(action='ignore', category=ConvergenceWarning)
         pd.options.display.float_format = '{:,.4f}'.format
 
-        print(f'\nCorrelation metric: {self.comparison_metric.__name__}')
-
-        basic_statistical = self.statistical_evaluation()
+        basic_statistical = self.basic_statistical_evaluation()
         correlation_correlation = self.correlation_correlation()
         column_correlation = self.column_correlations()
-        estimators = self.estimator_evaluation(target_col=target_col, target_type=target_type)
-        pca_variance = self.pca_correlation()
+        estimators = self.estimator_evaluation(target_col=target_col, target_type=target_type, kfold=kfold)
         nearest_neighbor = self.row_distance(n_samples=n_samples_distance)
 
-        miscellaneous = {}
-        miscellaneous['Column Correlation Distance RMSE'] = self.correlation_distance(how='rmse')
-        miscellaneous['Column Correlation distance MAE'] = self.correlation_distance(how='mae')
+        miscellaneous_dict = {
+            'Column Correlation Distance RMSE': self.correlation_distance(how='rmse'),
+            'Column Correlation distance MAE': self.correlation_distance(how='mae'),
+        }
 
-        miscellaneous['Duplicate rows between sets (real/fake)'] = self.get_duplicates()
-        miscellaneous['nearest neighbor mean'] = nearest_neighbor[0]
-        miscellaneous['nearest neighbor std'] = nearest_neighbor[1]
-        miscellaneous_df = pd.DataFrame({'Result': list(miscellaneous.values())}, index=list(miscellaneous.keys()))
-        print(f'\nMiscellaneous results:')
-        print(miscellaneous_df.to_string())
+        miscellaneous = pd.DataFrame({'Result': list(miscellaneous_dict.values())},
+                                     index=list(miscellaneous_dict.keys()))
 
-        all_results = {
+        privacy_metrics_dict = {
+            'Duplicate rows between sets (real/fake)': self.get_duplicates(),
+            'nearest neighbor mean': nearest_neighbor[0],
+            'nearest neighbor std': nearest_neighbor[1],
+        }
+
+        privacy_report = EvaluationResult(
+            name='Privacy Results',
+            content=dict_to_df(privacy_metrics_dict),
+        )
+
+        privacy_tab = [privacy_report]
+
+
+        efficacy_title = 'Classifier F1-scores and their Jaccard similarities:' if self.target_type == 'class' \
+            else '\nRegressor MSE-scores'
+
+
+        ml_efficacy_tab = [
+            EvaluationResult(name=efficacy_title, content=self.estimators_scores)
+        ]
+
+
+        js_df = js_distance_df(self.real, self.fake, self.numerical_columns)
+
+        statistical_tab = [
+            EvaluationResult(name='Jensen-Shannon distance', content=js_df,
+                             appendix=f'### Mean: {js_df.js_distance.mean(): .3f}'),
+            EvaluationResult(name='Kolmogorov-Smirnov statistic',
+                             content=kolmogorov_smirnov_df(self.real, self.fake, self.numerical_columns)
+                             )
+        ]
+
+
+        all_results_dict = {
             'Basic statistics': basic_statistical,
             'Correlation column correlations': correlation_correlation,
             'Mean Correlation between fake and real columns': column_correlation,
             f'{"1 - MAPE Estimator results" if self.target_type == "class" else "Correlation RMSE"}': estimators,
-            # '1 - MAPE 5 PCA components': pca_variance,
         }
-        total_result = np.mean(list(all_results.values()))
-        all_results['Similarity Score'] = total_result
-        all_results_df = pd.DataFrame({'Result': list(all_results.values())}, index=list(all_results.keys()))
+        all_results_dict['Similarity Score'] = np.mean(list(all_results_dict.values()))
 
-        print(f'\nResults:')
-        print(all_results_df.to_string())
-        return all_results
+        summary = EvaluationResult(
+            name='Overview Results',
+            content=dict_to_df(all_results_dict)
+        )
+
+        overview_tab = [summary, ]
+
+        if return_outputs:
+            all_results = [
+                *overview_tab,
+                *ml_efficacy_tab,
+                *privacy_tab,
+                *statistical_tab,
+            ]
+
+            all_results = {x.name: x.content.to_dict(orient='index') for x in all_results}
+
+            return all_results
+
+        if notebook:
+            visualize_notebook(
+                self,
+                overview=overview_tab,
+                privacy_metrics=privacy_tab,
+                ml_efficacy=ml_efficacy_tab,
+                statistical=statistical_tab,
+            )
+
+        else:
+            print(f'\n{efficacy_title}:')
+            print(self.estimators_scores.to_string())
+
+            print(f'\nPrivacy results:')
+            print(privacy_report.content.to_string())
+
+            print(f'\nMiscellaneous results:')
+            print(miscellaneous.to_string())
+
+            print(f'\nResults:')
+            print(summary.content.to_string())
